# Copyright 2021-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import dataclasses
import enum
import struct
from typing import TYPE_CHECKING, ClassVar

from typing_extensions import Self

from bumble import core, device, gatt, gatt_client, utils

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------


class PlayingOrder(utils.OpenIntEnum):
    '''See Media Control Service 3.15. Playing Order.'''

    SINGLE_ONCE = 0x01
    SINGLE_REPEAT = 0x02
    IN_ORDER_ONCE = 0x03
    IN_ORDER_REPEAT = 0x04
    OLDEST_ONCE = 0x05
    OLDEST_REPEAT = 0x06
    NEWEST_ONCE = 0x07
    NEWEST_REPEAT = 0x08
    SHUFFLE_ONCE = 0x09
    SHUFFLE_REPEAT = 0x0A


class PlayingOrderSupported(enum.IntFlag):
    '''See Media Control Service 3.16. Playing Orders Supported.'''

    SINGLE_ONCE = 0x0001
    SINGLE_REPEAT = 0x0002
    IN_ORDER_ONCE = 0x0004
    IN_ORDER_REPEAT = 0x0008
    OLDEST_ONCE = 0x0010
    OLDEST_REPEAT = 0x0020
    NEWEST_ONCE = 0x0040
    NEWEST_REPEAT = 0x0080
    SHUFFLE_ONCE = 0x0100
    SHUFFLE_REPEAT = 0x0200


class MediaState(utils.OpenIntEnum):
    '''See Media Control Service 3.17. Media State.'''

    INACTIVE = 0x00
    PLAYING = 0x01
    PAUSED = 0x02
    SEEKING = 0x03


class MediaControlPointOpcode(utils.OpenIntEnum):
    '''See Media Control Service 3.18. Media Control Point.'''

    PLAY = 0x01
    PAUSE = 0x02
    FAST_REWIND = 0x03
    FAST_FORWARD = 0x04
    STOP = 0x05
    MOVE_RELATIVE = 0x10
    PREVIOUS_SEGMENT = 0x20
    NEXT_SEGMENT = 0x21
    FIRST_SEGMENT = 0x22
    LAST_SEGMENT = 0x23
    GOTO_SEGMENT = 0x24
    PREVIOUS_TRACK = 0x30
    NEXT_TRACK = 0x31
    FIRST_TRACK = 0x32
    LAST_TRACK = 0x33
    GOTO_TRACK = 0x34
    PREVIOUS_GROUP = 0x40
    NEXT_GROUP = 0x41
    FIRST_GROUP = 0x42
    LAST_GROUP = 0x43
    GOTO_GROUP = 0x44


class MediaControlPointResultCode(enum.IntFlag):
    '''See Media Control Service 3.18.2. Media Control Point Notification.'''

    SUCCESS = 0x01
    OPCODE_NOT_SUPPORTED = 0x02
    MEDIA_PLAYER_INACTIVE = 0x03
    COMMAND_CANNOT_BE_COMPLETED = 0x04


class MediaControlPointOpcodeSupported(enum.IntFlag):
    '''See Media Control Service 3.19. Media Control Point Opcodes Supported.'''

    PLAY = 0x00000001
    PAUSE = 0x00000002
    FAST_REWIND = 0x00000004
    FAST_FORWARD = 0x00000008
    STOP = 0x00000010
    MOVE_RELATIVE = 0x00000020
    PREVIOUS_SEGMENT = 0x00000040
    NEXT_SEGMENT = 0x00000080
    FIRST_SEGMENT = 0x00000100
    LAST_SEGMENT = 0x00000200
    GOTO_SEGMENT = 0x00000400
    PREVIOUS_TRACK = 0x00000800
    NEXT_TRACK = 0x00001000
    FIRST_TRACK = 0x00002000
    LAST_TRACK = 0x00004000
    GOTO_TRACK = 0x00008000
    PREVIOUS_GROUP = 0x00010000
    NEXT_GROUP = 0x00020000
    FIRST_GROUP = 0x00040000
    LAST_GROUP = 0x00080000
    GOTO_GROUP = 0x00100000


class SearchControlPointItemType(utils.OpenIntEnum):
    '''See Media Control Service 3.20. Search Control Point.'''

    TRACK_NAME = 0x01
    ARTIST_NAME = 0x02
    ALBUM_NAME = 0x03
    GROUP_NAME = 0x04
    EARLIEST_YEAR = 0x05
    LATEST_YEAR = 0x06
    GENRE = 0x07
    ONLY_TRACKS = 0x08
    ONLY_GROUPS = 0x09


class ObjectType(utils.OpenIntEnum):
    '''See Media Control Service 4.4.1. Object Type field.'''

    TASK = 0
    GROUP = 1


# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------


class ObjectId(int):
    '''See Media Control Service 4.4.2. Object ID field.'''

    @classmethod
    def create_from_bytes(cls: type[Self], data: bytes) -> Self:
        return cls(int.from_bytes(data, byteorder='little', signed=False))

    def __bytes__(self) -> bytes:
        return self.to_bytes(6, 'little')


@dataclasses.dataclass
class GroupObjectType:
    '''See Media Control Service 4.4. Group Object Type.'''

    object_type: ObjectType
    object_id: ObjectId

    @classmethod
    def from_bytes(cls: type[Self], data: bytes) -> Self:
        return cls(
            object_type=ObjectType(data[0]),
            object_id=ObjectId.create_from_bytes(data[1:]),
        )

    def __bytes__(self) -> bytes:
        return bytes([self.object_type]) + bytes(self.object_id)


# -----------------------------------------------------------------------------
# Server
# -----------------------------------------------------------------------------
class MediaControlService(gatt.TemplateService):
    '''Media Control Service server implementation, only for testing currently.'''

    UUID = gatt.GATT_MEDIA_CONTROL_SERVICE

    def __init__(self, media_player_name: str | None = None) -> None:
        self.track_position = 0

        self.media_player_name_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_MEDIA_PLAYER_NAME_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=(media_player_name or 'Bumble Player').encode(),
        )
        self.track_changed_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_TRACK_CHANGED_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=b'',
        )
        self.track_title_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_TRACK_TITLE_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=b'',
        )
        self.track_duration_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_TRACK_DURATION_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=b'',
        )
        self.track_position_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_TRACK_POSITION_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.WRITE
            | gatt.Characteristic.Properties.WRITE_WITHOUT_RESPONSE
            | gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION
            | gatt.Characteristic.Permissions.WRITE_REQUIRES_ENCRYPTION,
            value=b'',
        )
        self.media_state_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_MEDIA_STATE_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=b'',
        )
        self.media_control_point_characteristic: gatt.Characteristic[bytes] = (
            gatt.Characteristic(
                uuid=gatt.GATT_MEDIA_CONTROL_POINT_CHARACTERISTIC,
                properties=gatt.Characteristic.Properties.WRITE
                | gatt.Characteristic.Properties.WRITE_WITHOUT_RESPONSE
                | gatt.Characteristic.Properties.NOTIFY,
                permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION
                | gatt.Characteristic.Permissions.WRITE_REQUIRES_ENCRYPTION,
                value=gatt.CharacteristicValue(write=self.on_media_control_point),
            )
        )
        self.media_control_point_opcodes_supported_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_MEDIA_CONTROL_POINT_OPCODES_SUPPORTED_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.NOTIFY,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=b'',
        )
        self.content_control_id_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_CONTENT_CONTROL_ID_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=b'',
        )

        super().__init__(
            [
                self.media_player_name_characteristic,
                self.track_changed_characteristic,
                self.track_title_characteristic,
                self.track_duration_characteristic,
                self.track_position_characteristic,
                self.media_state_characteristic,
                self.media_control_point_characteristic,
                self.media_control_point_opcodes_supported_characteristic,
                self.content_control_id_characteristic,
            ]
        )

    async def on_media_control_point(
        self, connection: device.Connection, data: bytes
    ) -> None:
        opcode = MediaControlPointOpcode(data[0])

        await connection.device.notify_subscriber(
            connection,
            self.media_control_point_characteristic,
            value=bytes([opcode, MediaControlPointResultCode.SUCCESS]),
        )


class GenericMediaControlService(MediaControlService):
    UUID = gatt.GATT_GENERIC_MEDIA_CONTROL_SERVICE


# -----------------------------------------------------------------------------
# Client
# -----------------------------------------------------------------------------
class MediaControlServiceProxy(
    gatt_client.ProfileServiceProxy, utils.CompositeEventEmitter
):
    SERVICE_CLASS = MediaControlService

    _CHARACTERISTICS: ClassVar[dict[str, core.UUID]] = {
        'media_player_name': gatt.GATT_MEDIA_PLAYER_NAME_CHARACTERISTIC,
        'media_player_icon_object_id': gatt.GATT_MEDIA_PLAYER_ICON_OBJECT_ID_CHARACTERISTIC,
        'media_player_icon_url': gatt.GATT_MEDIA_PLAYER_ICON_URL_CHARACTERISTIC,
        'track_changed': gatt.GATT_TRACK_CHANGED_CHARACTERISTIC,
        'track_title': gatt.GATT_TRACK_TITLE_CHARACTERISTIC,
        'track_duration': gatt.GATT_TRACK_DURATION_CHARACTERISTIC,
        'track_position': gatt.GATT_TRACK_POSITION_CHARACTERISTIC,
        'playback_speed': gatt.GATT_PLAYBACK_SPEED_CHARACTERISTIC,
        'seeking_speed': gatt.GATT_SEEKING_SPEED_CHARACTERISTIC,
        'current_track_segments_object_id': gatt.GATT_CURRENT_TRACK_SEGMENTS_OBJECT_ID_CHARACTERISTIC,
        'current_track_object_id': gatt.GATT_CURRENT_TRACK_OBJECT_ID_CHARACTERISTIC,
        'next_track_object_id': gatt.GATT_NEXT_TRACK_OBJECT_ID_CHARACTERISTIC,
        'parent_group_object_id': gatt.GATT_PARENT_GROUP_OBJECT_ID_CHARACTERISTIC,
        'current_group_object_id': gatt.GATT_CURRENT_GROUP_OBJECT_ID_CHARACTERISTIC,
        'playing_order': gatt.GATT_PLAYING_ORDER_CHARACTERISTIC,
        'playing_orders_supported': gatt.GATT_PLAYING_ORDERS_SUPPORTED_CHARACTERISTIC,
        'media_state': gatt.GATT_MEDIA_STATE_CHARACTERISTIC,
        'media_control_point': gatt.GATT_MEDIA_CONTROL_POINT_CHARACTERISTIC,
        'media_control_point_opcodes_supported': gatt.GATT_MEDIA_CONTROL_POINT_OPCODES_SUPPORTED_CHARACTERISTIC,
        'search_control_point': gatt.GATT_SEARCH_CONTROL_POINT_CHARACTERISTIC,
        'search_results_object_id': gatt.GATT_SEARCH_RESULTS_OBJECT_ID_CHARACTERISTIC,
        'content_control_id': gatt.GATT_CONTENT_CONTROL_ID_CHARACTERISTIC,
    }

    EVENT_MEDIA_STATE = "media_state"
    EVENT_TRACK_CHANGED = "track_changed"
    EVENT_TRACK_TITLE = "track_title"
    EVENT_TRACK_DURATION = "track_duration"
    EVENT_TRACK_POSITION = "track_position"

    media_player_name: gatt_client.CharacteristicProxy[bytes] | None = None
    media_player_icon_object_id: gatt_client.CharacteristicProxy[bytes] | None = None
    media_player_icon_url: gatt_client.CharacteristicProxy[bytes] | None = None
    track_changed: gatt_client.CharacteristicProxy[bytes] | None = None
    track_title: gatt_client.CharacteristicProxy[bytes] | None = None
    track_duration: gatt_client.CharacteristicProxy[bytes] | None = None
    track_position: gatt_client.CharacteristicProxy[bytes] | None = None
    playback_speed: gatt_client.CharacteristicProxy[bytes] | None = None
    seeking_speed: gatt_client.CharacteristicProxy[bytes] | None = None
    current_track_segments_object_id: gatt_client.CharacteristicProxy[bytes] | None = (
        None
    )
    current_track_object_id: gatt_client.CharacteristicProxy[bytes] | None = None
    next_track_object_id: gatt_client.CharacteristicProxy[bytes] | None = None
    parent_group_object_id: gatt_client.CharacteristicProxy[bytes] | None = None
    current_group_object_id: gatt_client.CharacteristicProxy[bytes] | None = None
    playing_order: gatt_client.CharacteristicProxy[bytes] | None = None
    playing_orders_supported: gatt_client.CharacteristicProxy[bytes] | None = None
    media_state: gatt_client.CharacteristicProxy[bytes] | None = None
    media_control_point: gatt_client.CharacteristicProxy[bytes] | None = None
    media_control_point_opcodes_supported: (
        gatt_client.CharacteristicProxy[bytes] | None
    ) = None
    search_control_point: gatt_client.CharacteristicProxy[bytes] | None = None
    search_results_object_id: gatt_client.CharacteristicProxy[bytes] | None = None
    content_control_id: gatt_client.CharacteristicProxy[bytes] | None = None

    if TYPE_CHECKING:
        media_control_point_notifications: asyncio.Queue[bytes]

    def __init__(self, service_proxy: gatt_client.ServiceProxy) -> None:
        utils.CompositeEventEmitter.__init__(self)
        self.service_proxy = service_proxy
        self.lock = asyncio.Lock()
        self.media_control_point_notifications = asyncio.Queue()

        for field, uuid in self._CHARACTERISTICS.items():
            if characteristics := service_proxy.get_characteristics_by_uuid(uuid):
                setattr(self, field, characteristics[0])

    async def subscribe_characteristics(self) -> None:
        if self.media_control_point:
            await self.media_control_point.subscribe(self._on_media_control_point)
        if self.media_state:
            await self.media_state.subscribe(self._on_media_state)
        if self.track_changed:
            await self.track_changed.subscribe(self._on_track_changed)
        if self.track_title:
            await self.track_title.subscribe(self._on_track_title)
        if self.track_duration:
            await self.track_duration.subscribe(self._on_track_duration)
        if self.track_position:
            await self.track_position.subscribe(self._on_track_position)

    async def write_control_point(
        self, opcode: MediaControlPointOpcode
    ) -> MediaControlPointResultCode:
        '''Writes a Media Control Point Opcode to peer and waits for the notification.

        The write operation will be executed when there isn't other pending commands.

        Args:
            opcode: opcode defined in `MediaControlPointOpcode`.

        Returns:
            Response code provided in `MediaControlPointResultCode`

        Raises:
            InvalidOperationError: Server does not have Media Control Point Characteristic.
            InvalidStateError: Server replies a notification with mismatched opcode.
        '''
        if not self.media_control_point:
            raise core.InvalidOperationError("Peer does not have media control point")

        async with self.lock:
            await self.media_control_point.write_value(
                bytes([opcode]),
                with_response=False,
            )

            (
                response_opcode,
                response_code,
            ) = await self.media_control_point_notifications.get()
            if response_opcode != opcode:
                raise core.InvalidStateError(
                    f"Expected {opcode} notification, but get {response_opcode}"
                )
            return MediaControlPointResultCode(response_code)

    def _on_media_control_point(self, data: bytes) -> None:
        self.media_control_point_notifications.put_nowait(data)

    def _on_media_state(self, data: bytes) -> None:
        self.emit(self.EVENT_MEDIA_STATE, MediaState(data[0]))

    def _on_track_changed(self, data: bytes) -> None:
        del data
        self.emit(self.EVENT_TRACK_CHANGED)

    def _on_track_title(self, data: bytes) -> None:
        self.emit(self.EVENT_TRACK_TITLE, data.decode("utf-8"))

    def _on_track_duration(self, data: bytes) -> None:
        self.emit(self.EVENT_TRACK_DURATION, struct.unpack_from('<i', data)[0])

    def _on_track_position(self, data: bytes) -> None:
        self.emit(self.EVENT_TRACK_POSITION, struct.unpack_from('<i', data)[0])


class GenericMediaControlServiceProxy(MediaControlServiceProxy):
    SERVICE_CLASS = GenericMediaControlService



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import functools
import logging

import pytest

from bumble import device
from bumble.hci import CodecID, CodingFormat
from bumble.profiles.ascs import (
    ASE_Config_Codec,
    ASE_Config_QOS,
    ASE_Disable,
    ASE_Enable,
    ASE_Operation,
    ASE_Receiver_Start_Ready,
    ASE_Receiver_Stop_Ready,
    ASE_Release,
    ASE_Update_Metadata,
    AseStateMachine,
    AudioStreamControlService,
    AudioStreamControlServiceProxy,
)
from bumble.profiles.bap import (
    AudioLocation,
    BasicAudioAnnouncement,
    BroadcastAudioAnnouncement,
    CodecSpecificCapabilities,
    CodecSpecificConfiguration,
    ContextType,
    FrameDuration,
    SamplingFrequency,
    SupportedFrameDuration,
    SupportedSamplingFrequency,
)
from bumble.profiles.le_audio import Metadata
from bumble.profiles.pacs import (
    PacRecord,
    PublishedAudioCapabilitiesService,
    PublishedAudioCapabilitiesServiceProxy,
)
from tests.test_utils import TwoDevices, async_barrier

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def basic_check(operation: ASE_Operation):
    serialized = bytes(operation)
    parsed = ASE_Operation.from_bytes(serialized)
    assert bytes(parsed) == serialized


# -----------------------------------------------------------------------------
def test_codec_specific_capabilities() -> None:
    SAMPLE_FREQUENCY = SupportedSamplingFrequency.FREQ_16000
    FRAME_SURATION = SupportedFrameDuration.DURATION_10000_US_SUPPORTED
    AUDIO_CHANNEL_COUNTS = [1]
    cap = CodecSpecificCapabilities(
        supported_sampling_frequencies=SAMPLE_FREQUENCY,
        supported_frame_durations=FRAME_SURATION,
        supported_audio_channel_count=AUDIO_CHANNEL_COUNTS,
        min_octets_per_codec_frame=40,
        max_octets_per_codec_frame=40,
        supported_max_codec_frames_per_sdu=1,
    )
    assert CodecSpecificCapabilities.from_bytes(bytes(cap)) == cap


# -----------------------------------------------------------------------------
def test_pac_record() -> None:
    SAMPLE_FREQUENCY = SupportedSamplingFrequency.FREQ_16000
    FRAME_SURATION = SupportedFrameDuration.DURATION_10000_US_SUPPORTED
    AUDIO_CHANNEL_COUNTS = [1]
    cap = CodecSpecificCapabilities(
        supported_sampling_frequencies=SAMPLE_FREQUENCY,
        supported_frame_durations=FRAME_SURATION,
        supported_audio_channel_count=AUDIO_CHANNEL_COUNTS,
        min_octets_per_codec_frame=40,
        max_octets_per_codec_frame=40,
        supported_max_codec_frames_per_sdu=1,
    )

    pac_record = PacRecord(
        coding_format=CodingFormat(CodecID.LC3),
        codec_specific_capabilities=cap,
        metadata=Metadata([Metadata.Entry(tag=Metadata.Tag.VENDOR_SPECIFIC, data=b'')]),
    )
    assert PacRecord.from_bytes(bytes(pac_record)) == pac_record


# -----------------------------------------------------------------------------
def test_vendor_specific_pac_record() -> None:
    # Vendor-Specific codec, Google, ID=0xFFFF. No capabilities and metadata.
    RAW_DATA = bytes.fromhex('ffe000ffff0000')
    assert bytes(PacRecord.from_bytes(RAW_DATA)) == RAW_DATA


# -----------------------------------------------------------------------------
def test_ASE_Config_Codec() -> None:
    operation = ASE_Config_Codec(
        ase_id=[1, 2],
        target_latency=[3, 4],
        target_phy=[5, 6],
        codec_id=[CodingFormat(CodecID.LC3), CodingFormat(CodecID.LC3)],
        codec_specific_configuration=[b'foo', b'bar'],
    )
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Config_QOS() -> None:
    operation = ASE_Config_QOS(
        ase_id=[1, 2],
        cig_id=[1, 2],
        cis_id=[3, 4],
        sdu_interval=[5, 6],
        framing=[0, 1],
        phy=[2, 3],
        max_sdu=[4, 5],
        retransmission_number=[6, 7],
        max_transport_latency=[8, 9],
        presentation_delay=[10, 11],
    )
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Enable() -> None:
    operation = ASE_Enable(
        ase_id=[1, 2],
        metadata=[b'', b''],
    )
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Update_Metadata() -> None:
    operation = ASE_Update_Metadata(
        ase_id=[1, 2],
        metadata=[b'', b''],
    )
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Disable() -> None:
    operation = ASE_Disable(ase_id=[1, 2])
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Release() -> None:
    operation = ASE_Release(ase_id=[1, 2])
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Receiver_Start_Ready() -> None:
    operation = ASE_Receiver_Start_Ready(ase_id=[1, 2])
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_ASE_Receiver_Stop_Ready() -> None:
    operation = ASE_Receiver_Stop_Ready(ase_id=[1, 2])
    basic_check(operation)


# -----------------------------------------------------------------------------
def test_codec_specific_configuration() -> None:
    SAMPLE_FREQUENCY = SamplingFrequency.FREQ_16000
    FRAME_SURATION = FrameDuration.DURATION_10000_US
    AUDIO_LOCATION = AudioLocation.FRONT_LEFT
    config = CodecSpecificConfiguration(
        sampling_frequency=SAMPLE_FREQUENCY,
        frame_duration=FRAME_SURATION,
        audio_channel_allocation=AUDIO_LOCATION,
        octets_per_codec_frame=60,
        codec_frames_per_sdu=1,
    )
    assert CodecSpecificConfiguration.from_bytes(bytes(config)) == config


# -----------------------------------------------------------------------------
def test_broadcast_audio_announcement() -> None:
    broadcast_audio_announcement = BroadcastAudioAnnouncement(123456)
    assert (
        BroadcastAudioAnnouncement.from_bytes(bytes(broadcast_audio_announcement))
        == broadcast_audio_announcement
    )


# -----------------------------------------------------------------------------
def test_basic_audio_announcement() -> None:
    basic_audio_announcement = BasicAudioAnnouncement(
        presentation_delay=40000,
        subgroups=[
            BasicAudioAnnouncement.Subgroup(
                codec_id=CodingFormat(codec_id=CodecID.LC3),
                codec_specific_configuration=CodecSpecificConfiguration(
                    sampling_frequency=SamplingFrequency.FREQ_48000,
                    frame_duration=FrameDuration.DURATION_10000_US,
                    octets_per_codec_frame=100,
                ),
                metadata=Metadata(
                    [
                        Metadata.Entry(tag=Metadata.Tag.LANGUAGE, data=b'eng'),
                        Metadata.Entry(tag=Metadata.Tag.PROGRAM_INFO, data=b'Disco'),
                    ]
                ),
                bis=[
                    BasicAudioAnnouncement.BIS(
                        index=0,
                        codec_specific_configuration=CodecSpecificConfiguration(
                            audio_channel_allocation=AudioLocation.FRONT_LEFT
                        ),
                    ),
                    BasicAudioAnnouncement.BIS(
                        index=1,
                        codec_specific_configuration=CodecSpecificConfiguration(
                            audio_channel_allocation=AudioLocation.FRONT_RIGHT
                        ),
                    ),
                ],
            )
        ],
    )
    assert (
        BasicAudioAnnouncement.from_bytes(bytes(basic_audio_announcement))
        == basic_audio_announcement
    )


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_pacs():
    devices = TwoDevices()
    devices[0].add_service(
        PublishedAudioCapabilitiesService(
            supported_sink_context=ContextType.MEDIA,
            available_sink_context=ContextType.MEDIA,
            supported_source_context=0,
            available_source_context=0,
            sink_pac=[
                # Codec Capability Setting 16_2
                PacRecord(
                    coding_format=CodingFormat(CodecID.LC3),
                    codec_specific_capabilities=CodecSpecificCapabilities(
                        supported_sampling_frequencies=(
                            SupportedSamplingFrequency.FREQ_16000
                        ),
                        supported_frame_durations=(
                            SupportedFrameDuration.DURATION_10000_US_SUPPORTED
                        ),
                        supported_audio_channel_count=[1],
                        min_octets_per_codec_frame=40,
                        max_octets_per_codec_frame=40,
                        supported_max_codec_frames_per_sdu=1,
                    ),
                ),
                # Codec Capability Setting 24_2
                PacRecord(
                    coding_format=CodingFormat(CodecID.LC3),
                    codec_specific_capabilities=CodecSpecificCapabilities(
                        supported_sampling_frequencies=(
                            SupportedSamplingFrequency.FREQ_24000
                        ),
                        supported_frame_durations=(
                            SupportedFrameDuration.DURATION_10000_US_SUPPORTED
                        ),
                        supported_audio_channel_count=[1],
                        min_octets_per_codec_frame=60,
                        max_octets_per_codec_frame=60,
                        supported_max_codec_frames_per_sdu=1,
                    ),
                ),
            ],
            sink_audio_locations=AudioLocation.FRONT_LEFT | AudioLocation.FRONT_RIGHT,
        )
    )

    await devices.setup_connection()
    peer = device.Peer(devices.connections[1])
    await peer.discover_service_and_create_proxy(PublishedAudioCapabilitiesServiceProxy)


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_ascs():
    devices = TwoDevices()
    devices[1].add_service(
        AudioStreamControlService(device=devices[1], sink_ase_id=[1, 2])
    )

    await devices.setup_connection()
    peer = device.Peer(devices.connections[0])
    ascs_client = await peer.discover_service_and_create_proxy(
        AudioStreamControlServiceProxy
    )

    notifications = {1: asyncio.Queue(), 2: asyncio.Queue()}

    def on_notification(data: bytes, ase_id: int):
        notifications[ase_id].put_nowait(data)

    # Should be idle
    assert await ascs_client.sink_ase[0].read_value() == bytes(
        [1, AseStateMachine.State.IDLE]
    )
    assert await ascs_client.sink_ase[1].read_value() == bytes(
        [2, AseStateMachine.State.IDLE]
    )

    # Subscribe
    await ascs_client.sink_ase[0].subscribe(
        functools.partial(on_notification, ase_id=1)
    )
    await ascs_client.sink_ase[1].subscribe(
        functools.partial(on_notification, ase_id=2)
    )

    # Config Codec
    config = CodecSpecificConfiguration(
        sampling_frequency=SamplingFrequency.FREQ_48000,
        frame_duration=FrameDuration.DURATION_10000_US,
        audio_channel_allocation=AudioLocation.FRONT_LEFT,
        octets_per_codec_frame=120,
        codec_frames_per_sdu=1,
    )
    await ascs_client.ase_control_point.write_value(
        ASE_Config_Codec(
            ase_id=[1, 2],
            target_latency=[3, 4],
            target_phy=[5, 6],
            codec_id=[CodingFormat(CodecID.LC3), CodingFormat(CodecID.LC3)],
            codec_specific_configuration=[config, config],
        )
    )
    assert (await notifications[1].get())[:2] == bytes(
        [1, AseStateMachine.State.CODEC_CONFIGURED]
    )
    assert (await notifications[2].get())[:2] == bytes(
        [2, AseStateMachine.State.CODEC_CONFIGURED]
    )

    # Config QOS
    await ascs_client.ase_control_point.write_value(
        ASE_Config_QOS(
            ase_id=[1, 2],
            cig_id=[1, 1],
            cis_id=[3, 4],
            sdu_interval=[5, 6],
            framing=[0, 1],
            phy=[2, 3],
            max_sdu=[4, 5],
            retransmission_number=[6, 7],
            max_transport_latency=[8, 9],
            presentation_delay=[10, 11],
        )
    )
    assert (await notifications[1].get())[:2] == bytes(
        [1, AseStateMachine.State.QOS_CONFIGURED]
    )
    assert (await notifications[2].get())[:2] == bytes(
        [2, AseStateMachine.State.QOS_CONFIGURED]
    )

    # Enable
    await ascs_client.ase_control_point.write_value(
        ASE_Enable(
            ase_id=[1, 2],
            metadata=[b'foo', b'bar'],
        )
    )
    assert (await notifications[1].get())[:2] == bytes(
        [1, AseStateMachine.State.ENABLING]
    )
    assert (await notifications[2].get())[:2] == bytes(
        [2, AseStateMachine.State.ENABLING]
    )

    # CIS establishment
    cis_handles = await devices[0].setup_cig(
        device.CigParameters(
            cig_id=1,
            cis_parameters=[
                device.CigParameters.CisParameters(cis_id=3),
                device.CigParameters.CisParameters(cis_id=4),
            ],
            sdu_interval_c_to_p=0,
            sdu_interval_p_to_c=0,
        )
    )
    await devices[0].create_cis(
        [(cis_handle, devices.connections[0]) for cis_handle in cis_handles]
    )
    assert (await notifications[1].get())[:2] == bytes(
        [1, AseStateMachine.State.STREAMING]
    )
    assert (await notifications[2].get())[:2] == bytes(
        [2, AseStateMachine.State.STREAMING]
    )

    # Release
    await ascs_client.ase_control_point.write_value(ASE_Release(ase_id=[1, 2]))
    assert (await notifications[1].get())[:2] == bytes(
        [1, AseStateMachine.State.RELEASING]
    )
    assert (await notifications[2].get())[:2] == bytes(
        [2, AseStateMachine.State.RELEASING]
    )
    assert (await notifications[1].get())[:2] == bytes([1, AseStateMachine.State.IDLE])
    assert (await notifications[2].get())[:2] == bytes([2, AseStateMachine.State.IDLE])

    await async_barrier()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_ascs_enable_source_then_sink():
    devices = TwoDevices()
    ascs_server = AudioStreamControlService(
        device=devices[1], sink_ase_id=[1], source_ase_id=[2]
    )
    sink_ase = ascs_server.ase_state_machines[1]
    source_ase = ascs_server.ase_state_machines[2]
    devices[1].add_service(ascs_server)
    condition = asyncio.Condition()

    async def on_state_change():
        async with condition:
            condition.notify_all()

    sink_ase.on(sink_ase.EVENT_STATE_CHANGE, on_state_change)
    source_ase.on(sink_ase.EVENT_STATE_CHANGE, on_state_change)

    await devices.setup_connection()
    peer = device.Peer(devices.connections[0])
    ascs_client = await peer.discover_service_and_create_proxy(
        AudioStreamControlServiceProxy
    )

    # Config Codec
    config = CodecSpecificConfiguration(
        sampling_frequency=SamplingFrequency.FREQ_48000,
        frame_duration=FrameDuration.DURATION_10000_US,
        audio_channel_allocation=AudioLocation.FRONT_LEFT,
        octets_per_codec_frame=120,
        codec_frames_per_sdu=1,
    )
    await ascs_client.ase_control_point.write_value(
        ASE_Config_Codec(
            ase_id=[1, 2],
            target_latency=[3, 4],
            target_phy=[5, 6],
            codec_id=[CodingFormat(CodecID.LC3), CodingFormat(CodecID.LC3)],
            codec_specific_configuration=[config, config],
        )
    )
    async with condition:
        await condition.wait_for(
            lambda: (
                sink_ase.state == AseStateMachine.State.CODEC_CONFIGURED
                and source_ase.state == AseStateMachine.State.CODEC_CONFIGURED
            )
        )

    # Config QOS
    await ascs_client.ase_control_point.write_value(
        ASE_Config_QOS(
            ase_id=[1, 2],
            cig_id=[1, 1],
            cis_id=[1, 1],
            sdu_interval=[100, 100],
            framing=[0, 0],
            phy=[1, 1],
            max_sdu=[100, 100],
            retransmission_number=[16, 16],
            max_transport_latency=[150, 150],
            presentation_delay=[10, 10],
        )
    )
    async with condition:
        await condition.wait_for(
            lambda: (
                sink_ase.state == AseStateMachine.State.QOS_CONFIGURED
                and source_ase.state == AseStateMachine.State.QOS_CONFIGURED
            )
        )

    # Enable ASE 2
    await ascs_client.ase_control_point.write_value(
        ASE_Enable(ase_id=[2], metadata=[b'foo'])
    )
    await async_barrier()
    cis_handles = await devices[0].setup_cig(
        device.CigParameters(
            cig_id=1,
            cis_parameters=[device.CigParameters.CisParameters(cis_id=1)],
            sdu_interval_c_to_p=100,
            sdu_interval_p_to_c=100,
        )
    )
    await devices[0].create_cis([(cis_handles[0], devices.connections[0])])

    async with condition:
        await condition.wait_for(
            lambda: (source_ase.state == AseStateMachine.State.ENABLING)
        )
    await ascs_client.ase_control_point.write_value(
        ASE_Receiver_Start_Ready(ase_id=[2])
    )
    async with condition:
        await condition.wait_for(
            lambda: (source_ase.state == AseStateMachine.State.STREAMING)
        )

    # Enable ASE 1
    await ascs_client.ase_control_point.write_value(
        ASE_Enable(ase_id=[1], metadata=[b'bar'])
    )
    async with condition:
        await condition.wait_for(
            lambda: (sink_ase.state == AseStateMachine.State.STREAMING)
        )



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import asyncio.subprocess
import enum
import json
import logging
import pathlib
import subprocess
import weakref
from importlib import resources

import aiohttp
import click
from aiohttp import web

import bumble
import bumble.logging
from bumble.a2dp import (
    A2DP_MPEG_2_4_AAC_CODEC_TYPE,
    A2DP_NON_A2DP_CODEC_TYPE,
    A2DP_SBC_CODEC_TYPE,
    AacMediaCodecInformation,
    OpusMediaCodecInformation,
    SbcMediaCodecInformation,
    make_audio_sink_service_sdp_records,
)
from bumble.avdtp import (
    AVDTP_AUDIO_MEDIA_TYPE,
    Listener,
    MediaCodecCapabilities,
    Protocol,
)
from bumble.codecs import AacAudioRtpPacket
from bumble.colors import color
from bumble.core import CommandTimeoutError, PhysicalTransport
from bumble.device import Connection, Device, DeviceConfiguration
from bumble.hci import HCI_StatusError
from bumble.pairing import PairingConfig
from bumble.rtp import MediaPacket
from bumble.sdp import ServiceAttribute
from bumble.transport import open_transport
from bumble.utils import AsyncRunner

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
DEFAULT_UI_PORT = 7654


# -----------------------------------------------------------------------------
class AudioExtractor:
    @staticmethod
    def create(codec: str):
        if codec == 'aac':
            return AacAudioExtractor()
        if codec == 'sbc':
            return SbcAudioExtractor()
        if codec == 'opus':
            return OpusAudioExtractor()

    def extract_audio(self, packet: MediaPacket) -> bytes:
        raise NotImplementedError()


# -----------------------------------------------------------------------------
class AacAudioExtractor:
    def extract_audio(self, packet: MediaPacket) -> bytes:
        return AacAudioRtpPacket.from_bytes(packet.payload).to_adts()


# -----------------------------------------------------------------------------
class SbcAudioExtractor:
    def extract_audio(self, packet: MediaPacket) -> bytes:
        # header = packet.payload[0]
        # fragmented = header >> 7
        # start = (header >> 6) & 0x01
        # last = (header >> 5) & 0x01
        # number_of_frames = header & 0x0F

        # TODO: support fragmented payloads
        return packet.payload[1:]


# -----------------------------------------------------------------------------
class OpusAudioExtractor:
    def extract_audio(self, packet: MediaPacket) -> bytes:
        # TODO: parse fields
        return packet.payload[1:]


# -----------------------------------------------------------------------------
class Output:
    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass

    async def suspend(self) -> None:
        pass

    async def on_connection(self, connection: Connection) -> None:
        pass

    async def on_disconnection(self, reason: int) -> None:
        pass

    def on_rtp_packet(self, packet: MediaPacket) -> None:
        pass


# -----------------------------------------------------------------------------
class FileOutput(Output):
    filename: str
    codec: str
    extractor: AudioExtractor

    def __init__(self, filename, codec):
        self.filename = filename
        self.codec = codec
        self.file = open(filename, 'wb')
        self.extractor = AudioExtractor.create(codec)

    def on_rtp_packet(self, packet: MediaPacket) -> None:
        self.file.write(self.extractor.extract_audio(packet))


# -----------------------------------------------------------------------------
class QueuedOutput(Output):
    MAX_QUEUE_SIZE = 32768

    packets: asyncio.Queue
    extractor: AudioExtractor
    packet_pump_task: asyncio.Task | None
    started: bool

    def __init__(self, extractor):
        self.extractor = extractor
        self.packets = asyncio.Queue()
        self.packet_pump_task = None
        self.started = False

    async def start(self):
        if self.started:
            return

        self.packet_pump_task = asyncio.create_task(self.pump_packets())

    async def pump_packets(self):
        while True:
            packet = await self.packets.get()
            await self.on_audio_packet(packet)

    async def on_audio_packet(self, packet: bytes) -> None:
        pass

    def on_rtp_packet(self, packet: MediaPacket) -> None:
        if self.packets.qsize() > self.MAX_QUEUE_SIZE:
            logger.debug("queue full, dropping")
            return

        self.packets.put_nowait(self.extractor.extract_audio(packet))


# -----------------------------------------------------------------------------
class WebSocketOutput(QueuedOutput):
    def __init__(self, codec, send_audio, send_message):
        super().__init__(AudioExtractor.create(codec))
        self.send_audio = send_audio
        self.send_message = send_message

    async def on_connection(self, connection: Connection) -> None:
        try:
            await connection.request_remote_name()
        except HCI_StatusError:
            pass
        peer_name = '' if connection.peer_name is None else connection.peer_name
        peer_address = connection.peer_address.to_string(False)
        await self.send_message(
            'connection',
            peer_address=peer_address,
            peer_name=peer_name,
        )

    async def on_disconnection(self, reason) -> None:
        await self.send_message('disconnection')

    async def on_audio_packet(self, packet: bytes) -> None:
        await self.send_audio(packet)

    async def start(self):
        await super().start()
        await self.send_message('start')

    async def stop(self):
        await super().stop()
        await self.send_message('stop')

    async def suspend(self):
        await super().suspend()
        await self.send_message('suspend')


# -----------------------------------------------------------------------------
class FfplayOutput(QueuedOutput):
    MAX_QUEUE_SIZE = 32768

    subprocess: asyncio.subprocess.Process | None
    ffplay_task: asyncio.Task | None

    def __init__(self, codec: str) -> None:
        super().__init__(AudioExtractor.create(codec))
        self.subprocess = None
        self.ffplay_task = None
        self.codec = codec

    async def start(self):
        if self.started:
            return

        await super().start()

        self.subprocess = await asyncio.create_subprocess_shell(
            f'ffplay -probesize 32 -f {self.codec} pipe:0',
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        self.ffplay_task = asyncio.create_task(self.monitor_ffplay())

    async def stop(self):
        # TODO
        pass

    async def suspend(self):
        # TODO
        pass

    async def monitor_ffplay(self):
        async def read_stream(name, stream):
            while True:
                data = await stream.read()
                logger.debug(f'{name}:', data)

        await asyncio.wait(
            [
                asyncio.create_task(
                    read_stream('[ffplay stdout]', self.subprocess.stdout)
                ),
                asyncio.create_task(
                    read_stream('[ffplay stderr]', self.subprocess.stderr)
                ),
                asyncio.create_task(self.subprocess.wait()),
            ]
        )
        logger.debug("FFPLAY done")

    async def on_audio_packet(self, packet):
        try:
            self.subprocess.stdin.write(packet)
        except Exception:
            logger.warning('!!!! exception while sending audio to ffplay pipe')


# -----------------------------------------------------------------------------
class UiServer:
    speaker: weakref.ReferenceType[Speaker]
    port: int

    def __init__(self, speaker: Speaker, port: int) -> None:
        self.speaker = weakref.ref(speaker)
        self.port = port
        self.channel_socket = None

    async def start_http(self) -> None:
        """Start the UI HTTP server."""

        app = web.Application()
        app.add_routes(
            [
                web.get('/', self.get_static),
                web.get('/speaker.html', self.get_static),
                web.get('/speaker.js', self.get_static),
                web.get('/speaker.css', self.get_static),
                web.get('/logo.svg', self.get_static),
                web.get('/channel', self.get_channel),
            ]
        )

        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, 'localhost', self.port)
        print('UI HTTP server at ' + color(f'http://127.0.0.1:{self.port}', 'green'))
        await site.start()

    async def get_static(self, request):
        path = request.path
        if path == '/':
            path = '/speaker.html'
        if path.endswith('.html'):
            content_type = 'text/html'
        elif path.endswith('.js'):
            content_type = 'text/javascript'
        elif path.endswith('.css'):
            content_type = 'text/css'
        elif path.endswith('.svg'):
            content_type = 'image/svg+xml'
        else:
            content_type = 'text/plain'
        text = (
            resources.files("bumble.apps.speaker")
            .joinpath(pathlib.Path(path).relative_to('/'))
            .read_text(encoding="utf-8")
        )
        return aiohttp.web.Response(text=text, content_type=content_type)

    async def get_channel(self, request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        # Process messages until the socket is closed.
        self.channel_socket = ws
        async for message in ws:
            if message.type == aiohttp.WSMsgType.TEXT:
                logger.debug(f'<<< received message: {message.data}')
                await self.on_message(message.data)
            elif message.type == aiohttp.WSMsgType.ERROR:
                logger.debug(
                    f'channel connection closed with exception {ws.exception()}'
                )

        self.channel_socket = None
        logger.debug('--- channel connection closed')

        return ws

    async def on_message(self, message_str: str):
        # Parse the message as JSON
        message = json.loads(message_str)

        # Dispatch the message
        message_type = message['type']
        message_params = message.get('params', {})
        handler = getattr(self, f'on_{message_type}_message')
        if handler:
            await handler(**message_params)

    async def on_hello_message(self):
        await self.send_message(
            'hello',
            bumble_version=bumble.__version__,
            codec=self.speaker().codec,
            streamState=self.speaker().stream_state.name,
        )
        if connection := self.speaker().connection:
            await self.send_message(
                'connection',
                peer_address=connection.peer_address.to_string(False),
                peer_name=connection.peer_name,
            )

    async def send_message(self, message_type: str, **kwargs) -> None:
        if self.channel_socket is None:
            return

        message = {'type': message_type, 'params': kwargs}
        await self.channel_socket.send_json(message)

    async def send_audio(self, data: bytes) -> None:
        if self.channel_socket is None:
            return

        try:
            await self.channel_socket.send_bytes(data)
        except Exception as error:
            logger.warning(f'exception while sending audio packet: {error}')


# -----------------------------------------------------------------------------
class Speaker:
    class StreamState(enum.Enum):
        IDLE = 0
        STOPPED = 1
        STARTED = 2
        SUSPENDED = 3

    def __init__(
        self,
        device_config,
        transport,
        codec,
        sampling_frequencies,
        bitrate,
        vbr,
        discover,
        outputs,
        ui_port,
    ):
        self.device_config = device_config
        self.transport = transport
        self.codec = codec
        self.sampling_frequencies = sampling_frequencies
        self.bitrate = bitrate
        self.vbr = vbr
        self.discover = discover
        self.ui_port = ui_port
        self.device = None
        self.connection = None
        self.listener = None
        self.packets_received = 0
        self.bytes_received = 0
        self.stream_state = Speaker.StreamState.IDLE
        self.outputs = []
        for output in outputs:
            if output == '@ffplay':
                self.outputs.append(FfplayOutput(codec))
                continue

            # Default to FileOutput
            self.outputs.append(FileOutput(output, codec))

        # Create an HTTP server for the UI
        self.ui_server = UiServer(speaker=self, port=ui_port)

    def sdp_records(self) -> dict[int, list[ServiceAttribute]]:
        service_record_handle = 0x00010001
        return {
            service_record_handle: make_audio_sink_service_sdp_records(
                service_record_handle
            )
        }

    def codec_capabilities(self) -> MediaCodecCapabilities:
        if self.codec == 'aac':
            return self.aac_codec_capabilities()

        if self.codec == 'sbc':
            return self.sbc_codec_capabilities()

        if self.codec == 'opus':
            return self.opus_codec_capabilities()

        raise RuntimeError('unsupported codec')

    def aac_codec_capabilities(self) -> MediaCodecCapabilities:
        supported_sampling_frequencies = AacMediaCodecInformation.SamplingFrequency(0)
        for sampling_frequency in self.sampling_frequencies or [
            8000,
            11025,
            12000,
            16000,
            22050,
            24000,
            32000,
            44100,
            48000,
        ]:
            supported_sampling_frequencies |= (
                AacMediaCodecInformation.SamplingFrequency.from_int(sampling_frequency)
            )
        return MediaCodecCapabilities(
            media_type=AVDTP_AUDIO_MEDIA_TYPE,
            media_codec_type=A2DP_MPEG_2_4_AAC_CODEC_TYPE,
            media_codec_information=AacMediaCodecInformation(
                object_type=AacMediaCodecInformation.ObjectType.MPEG_2_AAC_LC,
                sampling_frequency=supported_sampling_frequencies,
                channels=AacMediaCodecInformation.Channels.MONO
                | AacMediaCodecInformation.Channels.STEREO,
                vbr=1 if self.vbr else 0,
                bitrate=self.bitrate or 256000,
            ),
        )

    def sbc_codec_capabilities(self) -> MediaCodecCapabilities:
        supported_sampling_frequencies = SbcMediaCodecInformation.SamplingFrequency(0)
        for sampling_frequency in self.sampling_frequencies or [
            16000,
            32000,
            44100,
            48000,
        ]:
            supported_sampling_frequencies |= (
                SbcMediaCodecInformation.SamplingFrequency.from_int(sampling_frequency)
            )
        return MediaCodecCapabilities(
            media_type=AVDTP_AUDIO_MEDIA_TYPE,
            media_codec_type=A2DP_SBC_CODEC_TYPE,
            media_codec_information=SbcMediaCodecInformation(
                sampling_frequency=supported_sampling_frequencies,
                channel_mode=SbcMediaCodecInformation.ChannelMode.MONO
                | SbcMediaCodecInformation.ChannelMode.DUAL_CHANNEL
                | SbcMediaCodecInformation.ChannelMode.STEREO
                | SbcMediaCodecInformation.ChannelMode.JOINT_STEREO,
                block_length=SbcMediaCodecInformation.BlockLength.BL_4
                | SbcMediaCodecInformation.BlockLength.BL_8
                | SbcMediaCodecInformation.BlockLength.BL_12
                | SbcMediaCodecInformation.BlockLength.BL_16,
                subbands=SbcMediaCodecInformation.Subbands.S_4
                | SbcMediaCodecInformation.Subbands.S_8,
                allocation_method=SbcMediaCodecInformation.AllocationMethod.LOUDNESS
                | SbcMediaCodecInformation.AllocationMethod.SNR,
                minimum_bitpool_value=2,
                maximum_bitpool_value=53,
            ),
        )

    def opus_codec_capabilities(self) -> MediaCodecCapabilities:
        supported_sampling_frequencies = OpusMediaCodecInformation.SamplingFrequency(0)
        for sampling_frequency in self.sampling_frequencies or [48000]:
            supported_sampling_frequencies |= (
                OpusMediaCodecInformation.SamplingFrequency.from_int(sampling_frequency)
            )
        return MediaCodecCapabilities(
            media_type=AVDTP_AUDIO_MEDIA_TYPE,
            media_codec_type=A2DP_NON_A2DP_CODEC_TYPE,
            media_codec_information=OpusMediaCodecInformation(
                frame_size=OpusMediaCodecInformation.FrameSize.FS_10MS
                | OpusMediaCodecInformation.FrameSize.FS_20MS,
                channel_mode=OpusMediaCodecInformation.ChannelMode.MONO
                | OpusMediaCodecInformation.ChannelMode.STEREO
                | OpusMediaCodecInformation.ChannelMode.DUAL_MONO,
                sampling_frequency=supported_sampling_frequencies,
            ),
        )

    async def dispatch_to_outputs(self, function):
        for output in self.outputs:
            await function(output)

    def on_bluetooth_connection(self, connection):
        print(f'Connection: {connection}')
        self.connection = connection
        connection.on('disconnection', self.on_bluetooth_disconnection)
        AsyncRunner.spawn(
            self.dispatch_to_outputs(lambda output: output.on_connection(connection))
        )

    def on_bluetooth_disconnection(self, reason):
        print(f'Disconnection ({reason})')
        self.connection = None
        AsyncRunner.spawn(self.advertise())
        AsyncRunner.spawn(
            self.dispatch_to_outputs(lambda output: output.on_disconnection(reason))
        )

    def on_avdtp_connection(self, protocol):
        print('Audio Stream Open')

        # Add a sink endpoint to the server
        sink = protocol.add_sink(self.codec_capabilities())
        sink.on('start', self.on_sink_start)
        sink.on('stop', self.on_sink_stop)
        sink.on('suspend', self.on_sink_suspend)
        sink.on('configuration', lambda: self.on_sink_configuration(sink.configuration))
        sink.on('rtp_packet', self.on_rtp_packet)
        sink.on('rtp_channel_open', self.on_rtp_channel_open)
        sink.on('rtp_channel_close', self.on_rtp_channel_close)

        # Listen for close events
        protocol.on('close', self.on_avdtp_close)

        # Discover all endpoints on the remote device is requested
        if self.discover:
            AsyncRunner.spawn(self.discover_remote_endpoints(protocol))

    def on_avdtp_close(self):
        print("Audio Stream Closed")

    def on_sink_start(self):
        print("Sink Started\u001b[0K")
        self.stream_state = self.StreamState.STARTED
        AsyncRunner.spawn(self.dispatch_to_outputs(lambda output: output.start()))

    def on_sink_stop(self):
        print("Sink Stopped\u001b[0K")
        self.stream_state = self.StreamState.STOPPED
        AsyncRunner.spawn(self.dispatch_to_outputs(lambda output: output.stop()))

    def on_sink_suspend(self):
        print("Sink Suspended\u001b[0K")
        self.stream_state = self.StreamState.SUSPENDED
        AsyncRunner.spawn(self.dispatch_to_outputs(lambda output: output.suspend()))

    def on_sink_configuration(self, config):
        print("Sink Configuration:")
        print('\n'.join(["  " + str(capability) for capability in config]))

    def on_rtp_channel_open(self):
        print("RTP Channel Open")

    def on_rtp_channel_close(self):
        print("RTP Channel Closed")
        self.stream_state = self.StreamState.IDLE

    def on_rtp_packet(self, packet):
        self.packets_received += 1
        self.bytes_received += len(packet.payload)
        print(
            f'[{self.bytes_received} bytes in {self.packets_received} packets] {packet}',
            end='\r',
        )

        for output in self.outputs:
            output.on_rtp_packet(packet)

    async def advertise(self):
        await self.device.set_discoverable(True)
        await self.device.set_connectable(True)

    async def connect(self, address):
        # Connect to the source
        print(f'=== Connecting to {address}...')
        connection = await self.device.connect(
            address, transport=PhysicalTransport.BR_EDR
        )
        print(f'=== Connected to {connection.peer_address}')

        # Request authentication
        print('*** Authenticating...')
        await connection.authenticate()
        print('*** Authenticated')

        # Enable encryption
        print('*** Enabling encryption...')
        await connection.encrypt()
        print('*** Encryption on')

        protocol = await Protocol.connect(connection)
        self.listener.set_server(connection, protocol)
        self.on_avdtp_connection(protocol)

    async def discover_remote_endpoints(self, protocol):
        endpoints = await protocol.discover_remote_endpoints()
        print(f'@@@ Found {len(endpoints)} endpoints')
        for endpoint in endpoints:
            print('@@@', endpoint)

    async def run(self, connect_address):
        await self.ui_server.start_http()
        self.outputs.append(
            WebSocketOutput(
                self.codec, self.ui_server.send_audio, self.ui_server.send_message
            )
        )

        async with await open_transport(self.transport) as (hci_source, hci_sink):
            # Create a device
            device_config = DeviceConfiguration()
            if self.device_config:
                device_config.load_from_file(self.device_config)
            else:
                device_config.name = "Bumble Speaker"
                device_config.class_of_device = 0x240414
                device_config.keystore = "JsonKeyStore"

            device_config.classic_enabled = True
            device_config.le_enabled = False
            self.device = Device.from_config_with_hci(
                device_config, hci_source, hci_sink
            )

            # Setup the SDP to expose the sink service
            self.device.sdp_service_records = self.sdp_records()

            # Don't require MITM when pairing.
            self.device.pairing_config_factory = lambda connection: PairingConfig(
                mitm=False
            )

            # Start the controller
            await self.device.power_on()

            # Print some of the config/properties
            print("Speaker Name:", color(device_config.name, 'yellow'))
            print(
                "Speaker Bluetooth Address:",
                color(
                    self.device.public_address.to_string(with_type_qualifier=False),
                    'yellow',
                ),
            )

            # Listen for Bluetooth connections
            self.device.on('connection', self.on_bluetooth_connection)

            # Create a listener to wait for AVDTP connections
            self.listener = Listener.for_device(self.device)
            self.listener.on('connection', self.on_avdtp_connection)

            print(f'Speaker ready to play, codec={color(self.codec, "cyan")}')

            if connect_address:
                # Connect to the source
                try:
                    await self.connect(connect_address)
                except CommandTimeoutError:
                    print(color("Connection timed out", "red"))
                    return
            else:
                # Start being discoverable and connectable
                print("Waiting for connection...")
                await self.advertise()

            await hci_source.terminated

        for output in self.outputs:
            await output.stop()


# -----------------------------------------------------------------------------
@click.group()
@click.pass_context
def speaker_cli(ctx, device_config):
    ctx.ensure_object(dict)
    ctx.obj['device_config'] = device_config


@click.command()
@click.option(
    '--codec',
    type=click.Choice(['sbc', 'aac', 'opus']),
    default='aac',
    show_default=True,
)
@click.option(
    '--sampling-frequency',
    metavar='SAMPLING-FREQUENCY',
    type=int,
    multiple=True,
    help='Enable a sampling frequency (may be specified more than once)',
)
@click.option(
    '--bitrate',
    metavar='BITRATE',
    type=int,
    help='Supported bitrate (AAC only)',
)
@click.option(
    '--vbr/--no-vbr', is_flag=True, default=True, help='Enable VBR (AAC only)'
)
@click.option(
    '--discover', is_flag=True, help='Discover remote endpoints once connected'
)
@click.option(
    '--output',
    multiple=True,
    metavar='NAME',
    help=(
        'Send audio to this named output '
        '(may be used more than once for multiple outputs)'
    ),
)
@click.option(
    '--ui-port',
    'ui_port',
    metavar='HTTP_PORT',
    default=DEFAULT_UI_PORT,
    show_default=True,
    help='HTTP port for the UI server',
)
@click.option(
    '--connect',
    'connect_address',
    metavar='ADDRESS_OR_NAME',
    help='Address or name to connect to',
)
@click.option('--device-config', metavar='FILENAME', help='Device configuration file')
@click.argument('transport')
def speaker(
    transport,
    codec,
    sampling_frequency,
    bitrate,
    vbr,
    connect_address,
    discover,
    output,
    ui_port,
    device_config,
):
    """Run the speaker."""

    if '@ffplay' in output:
        # Check if ffplay is installed
        try:
            subprocess.run(['ffplay', '-version'], capture_output=True, check=True)
        except FileNotFoundError:
            print(
                color('ffplay not installed, @ffplay output will be disabled', 'yellow')
            )
            output = list(filter(lambda x: x != '@ffplay', output))

    asyncio.run(
        Speaker(
            device_config,
            transport,
            codec,
            sampling_frequency,
            bitrate,
            vbr,
            discover,
            output,
            ui_port,
        ).run(connect_address)
    )


# -----------------------------------------------------------------------------
def main():
    bumble.logging.setup_basic_logging('WARNING')
    speaker()


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic Access Profile"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import logging
import struct

from bumble.core import Appearance
from bumble.gatt import (
    GATT_APPEARANCE_CHARACTERISTIC,
    GATT_DEVICE_NAME_CHARACTERISTIC,
    GATT_GENERIC_ACCESS_SERVICE,
    Characteristic,
    TemplateService,
)
from bumble.gatt_adapters import (
    DelegatedCharacteristicProxyAdapter,
    UTF8CharacteristicProxyAdapter,
)
from bumble.gatt_client import CharacteristicProxy, ProfileServiceProxy, ServiceProxy

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
class GenericAccessService(TemplateService):
    UUID = GATT_GENERIC_ACCESS_SERVICE

    device_name_characteristic: Characteristic[bytes]
    appearance_characteristic: Characteristic[bytes]

    def __init__(
        self, device_name: str, appearance: Appearance | tuple[int, int] | int = 0
    ):
        match appearance:
            case int():
                appearance_int = appearance
            case tuple():
                appearance_int = (appearance[0] << 6) | appearance[1]
            case Appearance():
                appearance_int = int(appearance)
            case _:
                raise TypeError()

        self.device_name_characteristic = Characteristic(
            GATT_DEVICE_NAME_CHARACTERISTIC,
            Characteristic.Properties.READ,
            Characteristic.READABLE,
            device_name.encode('utf-8')[:248],
        )

        self.appearance_characteristic = Characteristic(
            GATT_APPEARANCE_CHARACTERISTIC,
            Characteristic.Properties.READ,
            Characteristic.READABLE,
            struct.pack('<H', appearance_int),
        )

        super().__init__(
            [self.device_name_characteristic, self.appearance_characteristic]
        )


# -----------------------------------------------------------------------------
class GenericAccessServiceProxy(ProfileServiceProxy):
    SERVICE_CLASS = GenericAccessService

    device_name: CharacteristicProxy[str] | None
    appearance: CharacteristicProxy[Appearance] | None

    def __init__(self, service_proxy: ServiceProxy):
        self.service_proxy = service_proxy

        if characteristics := service_proxy.get_characteristics_by_uuid(
            GATT_DEVICE_NAME_CHARACTERISTIC
        ):
            self.device_name = UTF8CharacteristicProxyAdapter(characteristics[0])
        else:
            self.device_name = None

        if characteristics := service_proxy.get_characteristics_by_uuid(
            GATT_APPEARANCE_CHARACTERISTIC
        ):
            self.appearance = DelegatedCharacteristicProxyAdapter(
                characteristics[0],
                decode=lambda value: Appearance.from_int(
                    struct.unpack_from('<H', value, 0)[0],
                ),
            )
        else:
            self.appearance = None



# Copyright 2021-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from bumble.profiles import le_audio


def test_parse_metadata():
    metadata = le_audio.Metadata(
        entries=[
            le_audio.Metadata.Entry(
                tag=le_audio.Metadata.Tag.PROGRAM_INFO,
                data=b'',
            ),
            le_audio.Metadata.Entry(
                tag=le_audio.Metadata.Tag.STREAMING_AUDIO_CONTEXTS,
                data=bytes([0, 0]),
            ),
            le_audio.Metadata.Entry(
                tag=le_audio.Metadata.Tag.PREFERRED_AUDIO_CONTEXTS,
                data=bytes([1, 2]),
            ),
        ]
    )

    assert le_audio.Metadata.from_bytes(bytes(metadata)) == metadata



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import contextlib
import logging
from collections.abc import AsyncGenerator, AsyncIterator, Awaitable, Callable
from typing import Any

import grpc
from google.protobuf import (
    any_pb2,  # pytype: disable=pyi-error
    empty_pb2,  # pytype: disable=pyi-error
    wrappers_pb2,  # pytype: disable=pyi-error
)
from pandora.host_pb2 import Connection
from pandora.security_grpc_aio import SecurityServicer, SecurityStorageServicer
from pandora.security_pb2 import (
    LE_LEVEL1,
    LE_LEVEL2,
    LE_LEVEL3,
    LE_LEVEL4,
    LEVEL0,
    LEVEL1,
    LEVEL2,
    LEVEL3,
    LEVEL4,
    DeleteBondRequest,
    IsBondedRequest,
    LESecurityLevel,
    PairingEvent,
    PairingEventAnswer,
    SecureRequest,
    SecureResponse,
    SecurityLevel,
    WaitSecurityRequest,
    WaitSecurityResponse,
)

import bumble.utils
from bumble import hci
from bumble.core import InvalidArgumentError, PhysicalTransport, ProtocolError
from bumble.device import Connection as BumbleConnection
from bumble.device import Device
from bumble.hci import HCI_Error, Role
from bumble.pairing import PairingConfig
from bumble.pairing import PairingDelegate as BasePairingDelegate
from bumble.pandora import utils
from bumble.pandora.config import Config


class PairingDelegate(BasePairingDelegate):
    def __init__(
        self,
        connection: BumbleConnection,
        service: SecurityService,
        io_capability: BasePairingDelegate.IoCapability = BasePairingDelegate.NO_OUTPUT_NO_INPUT,
        local_initiator_key_distribution: BasePairingDelegate.KeyDistribution = BasePairingDelegate.DEFAULT_KEY_DISTRIBUTION,
        local_responder_key_distribution: BasePairingDelegate.KeyDistribution = BasePairingDelegate.DEFAULT_KEY_DISTRIBUTION,
    ) -> None:
        self.log = utils.BumbleServerLoggerAdapter(
            logging.getLogger(),
            {'service_name': 'Security', 'device': connection.device},
        )
        self.connection = connection
        self.service = service
        super().__init__(
            io_capability,
            local_initiator_key_distribution,
            local_responder_key_distribution,
        )

    async def accept(self) -> bool:
        return True

    def add_origin(self, ev: PairingEvent) -> PairingEvent:
        if not self.connection.is_incomplete:
            assert ev.connection
            ev.connection.CopyFrom(
                Connection(
                    cookie=any_pb2.Any(value=self.connection.handle.to_bytes(4, 'big'))
                )
            )
        else:
            # In BR/EDR, connection may not be complete,
            # use address instead
            assert self.connection.transport == PhysicalTransport.BR_EDR
            ev.address = bytes(reversed(bytes(self.connection.peer_address)))

        return ev

    async def confirm(self, auto: bool = False) -> bool:
        self.log.debug(
            f"Pairing event: `just_works` (io_capability: {self.io_capability})"
        )

        if self.service.event_queue is None or self.service.event_answer is None:
            return True

        event = self.add_origin(PairingEvent(just_works=empty_pb2.Empty()))
        self.service.event_queue.put_nowait(event)
        answer = await anext(self.service.event_answer)  # type: ignore
        assert answer.event == event
        assert answer.answer_variant() == 'confirm' and answer.confirm is not None
        return answer.confirm

    async def compare_numbers(self, number: int, digits: int = 6) -> bool:
        self.log.debug(
            f"Pairing event: `numeric_comparison` (io_capability: {self.io_capability})"
        )

        if self.service.event_queue is None or self.service.event_answer is None:
            raise RuntimeError('security: unhandled number comparison request')

        event = self.add_origin(PairingEvent(numeric_comparison=number))
        self.service.event_queue.put_nowait(event)
        answer = await anext(self.service.event_answer)  # type: ignore
        assert answer.event == event
        assert answer.answer_variant() == 'confirm' and answer.confirm is not None
        return answer.confirm

    async def get_number(self) -> int | None:
        self.log.debug(
            f"Pairing event: `passkey_entry_request` (io_capability: {self.io_capability})"
        )

        if self.service.event_queue is None or self.service.event_answer is None:
            raise RuntimeError('security: unhandled number request')

        event = self.add_origin(PairingEvent(passkey_entry_request=empty_pb2.Empty()))
        self.service.event_queue.put_nowait(event)
        answer = await anext(self.service.event_answer)  # type: ignore
        assert answer.event == event
        if answer.answer_variant() is None:
            return None
        assert answer.answer_variant() == 'passkey'
        return answer.passkey

    async def get_string(self, max_length: int) -> str | None:
        self.log.debug(
            f"Pairing event: `pin_code_request` (io_capability: {self.io_capability})"
        )

        if self.service.event_queue is None or self.service.event_answer is None:
            raise RuntimeError('security: unhandled pin_code request')

        event = self.add_origin(PairingEvent(pin_code_request=empty_pb2.Empty()))
        self.service.event_queue.put_nowait(event)
        answer = await anext(self.service.event_answer)  # type: ignore
        assert answer.event == event
        if answer.answer_variant() is None:
            return None
        assert answer.answer_variant() == 'pin'

        if answer.pin is None:
            return None

        pin = answer.pin.decode('utf-8')
        if not pin or len(pin) > max_length:
            raise ValueError(f'Pin must be utf-8 encoded up to {max_length} bytes')

        return pin

    async def display_number(self, number: int, digits: int = 6) -> None:
        if (
            self.connection.transport == PhysicalTransport.BR_EDR
            and self.io_capability == BasePairingDelegate.DISPLAY_OUTPUT_ONLY
        ):
            return

        self.log.debug(
            f"Pairing event: `passkey_entry_notification` (io_capability: {self.io_capability})"
        )

        if self.service.event_queue is None:
            raise RuntimeError('security: unhandled number display request')

        event = self.add_origin(PairingEvent(passkey_entry_notification=number))
        self.service.event_queue.put_nowait(event)


class SecurityService(SecurityServicer):
    def __init__(self, device: Device, config: Config) -> None:
        self.log = utils.BumbleServerLoggerAdapter(
            logging.getLogger(), {'service_name': 'Security', 'device': device}
        )
        self.event_queue: asyncio.Queue[PairingEvent] | None = None
        self.event_answer: AsyncIterator[PairingEventAnswer] | None = None
        self.device = device
        self.config = config

        def pairing_config_factory(connection: BumbleConnection) -> PairingConfig:
            return PairingConfig(
                sc=config.pairing_sc_enable,
                mitm=config.pairing_mitm_enable,
                bonding=config.pairing_bonding_enable,
                identity_address_type=(
                    PairingConfig.AddressType.PUBLIC
                    if connection.self_address.is_public
                    else config.identity_address_type
                ),
                delegate=PairingDelegate(
                    connection,
                    self,
                    io_capability=config.io_capability,
                    local_initiator_key_distribution=config.smp_local_initiator_key_distribution,
                    local_responder_key_distribution=config.smp_local_responder_key_distribution,
                ),
            )

        self.device.pairing_config_factory = pairing_config_factory

    async def _classic_level_reached(
        self, level: SecurityLevel, connection: BumbleConnection
    ) -> bool:
        if level == LEVEL0:
            return True
        if level == LEVEL1:
            return connection.encryption == 0 or connection.authenticated
        if level == LEVEL2:
            return connection.encryption != 0 and connection.authenticated

        link_key_type: int | None = None
        if (keystore := connection.device.keystore) and (
            keys := await keystore.get(str(connection.peer_address))
        ):
            link_key_type = keys.link_key_type
        self.log.debug("link_key_type: %d", link_key_type)

        if level == LEVEL3:
            return (
                connection.encryption != 0
                and connection.authenticated
                and link_key_type
                in (
                    hci.LinkKeyType.AUTHENTICATED_COMBINATION_KEY_GENERATED_FROM_P_192,
                    hci.LinkKeyType.AUTHENTICATED_COMBINATION_KEY_GENERATED_FROM_P_256,
                )
            )
        if level == LEVEL4:
            return (
                connection.encryption == hci.HCI_Encryption_Change_Event.Enabled.AES_CCM
                and connection.authenticated
                and link_key_type
                == hci.LinkKeyType.AUTHENTICATED_COMBINATION_KEY_GENERATED_FROM_P_256
            )
        raise InvalidArgumentError(f"Unexpected level {level}")

    def _le_level_reached(
        self, level: LESecurityLevel, connection: BumbleConnection
    ) -> bool:
        if level == LE_LEVEL1:
            return True
        if level == LE_LEVEL2:
            return connection.encryption != 0
        if level == LE_LEVEL3:
            return connection.encryption != 0 and connection.authenticated
        if level == LE_LEVEL4:
            return (
                connection.encryption != 0
                and connection.authenticated
                and connection.sc
            )
        raise InvalidArgumentError(f"Unexpected level {level}")

    @utils.rpc
    async def OnPairing(
        self, request: AsyncIterator[PairingEventAnswer], context: grpc.ServicerContext
    ) -> AsyncGenerator[PairingEvent, None]:
        self.log.debug('OnPairing')

        if self.event_queue is not None:
            raise RuntimeError('already streaming pairing events')

        if len(self.device.connections):
            raise RuntimeError(
                'the `OnPairing` method shall be initiated before establishing any connections.'
            )

        self.event_queue = asyncio.Queue()
        self.event_answer = request

        try:
            while event := await self.event_queue.get():
                yield event

        finally:
            self.event_queue = None
            self.event_answer = None

    @utils.rpc
    async def Secure(
        self, request: SecureRequest, context: grpc.ServicerContext
    ) -> SecureResponse:
        connection_handle = int.from_bytes(request.connection.cookie.value, 'big')
        self.log.debug(f"Secure: {connection_handle}")

        connection = self.device.lookup_connection(connection_handle)
        assert connection

        oneof = request.WhichOneof('level')
        level = getattr(request, oneof)
        assert {PhysicalTransport.BR_EDR: 'classic', PhysicalTransport.LE: 'le'}[
            connection.transport
        ] == oneof

        # security level already reached
        if await self.reached_security_level(connection, level):
            return SecureResponse(success=empty_pb2.Empty())

        # trigger pairing if needed
        if self.need_pairing(connection, level):
            try:
                self.log.debug('Pair...')

                security_result = asyncio.get_running_loop().create_future()

                with contextlib.closing(bumble.utils.EventWatcher()) as watcher:

                    @watcher.on(connection, connection.EVENT_PAIRING)
                    def on_pairing(*_: Any) -> None:
                        security_result.set_result('success')

                    @watcher.on(connection, connection.EVENT_PAIRING_FAILURE)
                    def on_pairing_failure(*_: Any) -> None:
                        security_result.set_result('pairing_failure')

                    @watcher.on(connection, connection.EVENT_DISCONNECTION)
                    def on_disconnection(*_: Any) -> None:
                        security_result.set_result('connection_died')

                    if (
                        connection.transport == PhysicalTransport.LE
                        and connection.role == Role.PERIPHERAL
                    ):
                        connection.request_pairing()
                    else:
                        await connection.pair()

                    result = await security_result

                self.log.debug(f'Pairing session complete, status={result}')
                if result != 'success':
                    return SecureResponse(**{result: empty_pb2.Empty()})
            except asyncio.CancelledError:
                self.log.warning("Connection died during encryption")
                return SecureResponse(connection_died=empty_pb2.Empty())
            except (HCI_Error, ProtocolError) as e:
                self.log.warning(f"Pairing failure: {e}")
                return SecureResponse(pairing_failure=empty_pb2.Empty())

        # trigger authentication if needed
        if self.need_authentication(connection, level):
            try:
                self.log.debug('Authenticate...')
                await connection.authenticate()
                self.log.debug('Authenticated')
            except asyncio.CancelledError:
                self.log.warning("Connection died during authentication")
                return SecureResponse(connection_died=empty_pb2.Empty())
            except (HCI_Error, ProtocolError) as e:
                self.log.warning(f"Authentication failure: {e}")
                return SecureResponse(authentication_failure=empty_pb2.Empty())

        # trigger encryption if needed
        if self.need_encryption(connection, level):
            try:
                self.log.debug('Encrypt...')
                await connection.encrypt()
                self.log.debug('Encrypted')
            except asyncio.CancelledError:
                self.log.warning("Connection died during encryption")
                return SecureResponse(connection_died=empty_pb2.Empty())
            except (HCI_Error, ProtocolError) as e:
                self.log.warning(f"Encryption failure: {e}")
                return SecureResponse(encryption_failure=empty_pb2.Empty())

        # security level has been reached ?
        if await self.reached_security_level(connection, level):
            return SecureResponse(success=empty_pb2.Empty())
        return SecureResponse(not_reached=empty_pb2.Empty())

    @utils.rpc
    async def WaitSecurity(
        self, request: WaitSecurityRequest, context: grpc.ServicerContext
    ) -> WaitSecurityResponse:
        connection_handle = int.from_bytes(request.connection.cookie.value, 'big')
        self.log.debug(f"WaitSecurity: {connection_handle}")

        connection = self.device.lookup_connection(connection_handle)
        assert connection

        assert request.level
        level = request.level
        assert {PhysicalTransport.BR_EDR: 'classic', PhysicalTransport.LE: 'le'}[
            connection.transport
        ] == request.level_variant()

        wait_for_security: asyncio.Future[str] = (
            asyncio.get_running_loop().create_future()
        )
        authenticate_task: asyncio.Future[None] | None = None
        pair_task: asyncio.Future[None] | None = None

        async def authenticate() -> None:
            if (encryption := connection.encryption) != 0:
                self.log.debug('Disable encryption...')
                with contextlib.suppress(Exception):
                    await connection.encrypt(enable=False)
                self.log.debug('Disable encryption: done')

            self.log.debug('Authenticate...')
            await connection.authenticate()
            self.log.debug('Authenticate: done')

            if encryption != 0 and connection.encryption != encryption:
                self.log.debug('Re-enable encryption...')
                await connection.encrypt()
                self.log.debug('Re-enable encryption: done')

        def set_failure(name: str) -> Callable[..., None]:
            def wrapper(*args: Any) -> None:
                self.log.debug(f'Wait for security: error `{name}`: {args}')
                wait_for_security.set_result(name)

            return wrapper

        async def try_set_success(*_: Any) -> None:
            if await self.reached_security_level(connection, level):
                self.log.debug('Wait for security: done')
                wait_for_security.set_result('success')

        async def on_encryption_change(*_: Any) -> None:
            if await self.reached_security_level(connection, level):
                self.log.debug('Wait for security: done')
                wait_for_security.set_result('success')
            elif (
                connection.transport == PhysicalTransport.BR_EDR
                and self.need_authentication(connection, level)
            ):
                nonlocal authenticate_task
                if authenticate_task is None:
                    authenticate_task = asyncio.create_task(authenticate())

        def pair(*_: Any) -> None:
            if self.need_pairing(connection, level):
                bumble.utils.AsyncRunner.spawn(connection.pair())

        listeners: dict[str, Callable[..., None | Awaitable[None]]] = {
            'disconnection': set_failure('connection_died'),
            'pairing_failure': set_failure('pairing_failure'),
            'connection_authentication_failure': set_failure('authentication_failure'),
            'connection_encryption_failure': set_failure('encryption_failure'),
            'pairing': try_set_success,
            'connection_authentication': try_set_success,
            'connection_encryption_change': on_encryption_change,
            'classic_pairing': try_set_success,
            'classic_pairing_failure': set_failure('pairing_failure'),
            'security_request': pair,
        }

        with contextlib.closing(bumble.utils.EventWatcher()) as watcher:
            # register event handlers
            for event, listener in listeners.items():
                watcher.on(connection, event, listener)

            # security level already reached
            if await self.reached_security_level(connection, level):
                return WaitSecurityResponse(success=empty_pb2.Empty())

            self.log.debug('Wait for security...')
            kwargs = {}
            kwargs[await wait_for_security] = empty_pb2.Empty()

        # wait for `authenticate` to finish if any
        if authenticate_task is not None:
            self.log.debug('Wait for authentication...')
            with contextlib.suppress(Exception):
                await authenticate_task  # type: ignore
            self.log.debug('Authenticated')

        # wait for `pair` to finish if any
        if pair_task is not None:
            self.log.debug('Wait for authentication...')
            with contextlib.suppress(Exception):
                await pair_task  # type: ignore
            self.log.debug('paired')

        return WaitSecurityResponse(**kwargs)

    async def reached_security_level(
        self, connection: BumbleConnection, level: SecurityLevel | LESecurityLevel
    ) -> bool:
        self.log.debug(
            str(
                {
                    'level': level,
                    'encryption': connection.encryption,
                    'authenticated': connection.authenticated,
                    'sc': connection.sc,
                }
            )
        )

        if isinstance(level, LESecurityLevel):
            return self._le_level_reached(level, connection)

        return await self._classic_level_reached(level, connection)

    def need_pairing(self, connection: BumbleConnection, level: int) -> bool:
        if connection.transport == PhysicalTransport.LE:
            return level >= LE_LEVEL3 and not connection.authenticated
        return False

    def need_authentication(self, connection: BumbleConnection, level: int) -> bool:
        if connection.transport == PhysicalTransport.LE:
            return False
        if level == LEVEL2 and connection.encryption != 0:
            return not connection.authenticated
        return level >= LEVEL2 and not connection.authenticated

    def need_encryption(self, connection: BumbleConnection, level: int) -> bool:
        # TODO(abel): need to support MITM
        if connection.transport == PhysicalTransport.LE:
            return level == LE_LEVEL2 and not connection.encryption
        return level >= LEVEL2 and not connection.encryption


class SecurityStorageService(SecurityStorageServicer):
    def __init__(self, device: Device, config: Config) -> None:
        self.log = utils.BumbleServerLoggerAdapter(
            logging.getLogger(), {'service_name': 'SecurityStorage', 'device': device}
        )
        self.device = device
        self.config = config

    @utils.rpc
    async def IsBonded(
        self, request: IsBondedRequest, context: grpc.ServicerContext
    ) -> wrappers_pb2.BoolValue:
        address = utils.address_from_request(request, request.WhichOneof("address"))
        self.log.debug(f"IsBonded: {address}")

        if self.device.keystore is not None:
            is_bonded = await self.device.keystore.get(str(address)) is not None
        else:
            is_bonded = False

        return wrappers_pb2.BoolValue(value=is_bonded)

    @utils.rpc
    async def DeleteBond(
        self, request: DeleteBondRequest, context: grpc.ServicerContext
    ) -> empty_pb2.Empty:
        address = utils.address_from_request(request, request.WhichOneof("address"))
        self.log.debug(f"DeleteBond: {address}")

        if self.device.keystore is not None:
            with contextlib.suppress(KeyError):
                await self.device.keystore.delete(str(address))

        return empty_pb2.Empty()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_endpoint_description.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fgrpc_endpoint_description.proto\x12\x18\x61ndroid.emulation.remote\"V\n\x0b\x43redentials\x12\x16\n\x0epem_root_certs\x18\x01 \x01(\t\x12\x17\n\x0fpem_private_key\x18\x02 \x01(\t\x12\x16\n\x0epem_cert_chain\x18\x03 \x01(\t\"$\n\x06Header\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x96\x01\n\x08\x45ndpoint\x12\x0e\n\x06target\x18\x01 \x01(\t\x12>\n\x0ftls_credentials\x18\x02 \x01(\x0b\x32%.android.emulation.remote.Credentials\x12:\n\x10required_headers\x18\x03 \x03(\x0b\x32 .android.emulation.remote.HeaderB \n\x1c\x63om.android.emulation.remoteP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc_endpoint_description_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.android.emulation.remoteP\001'
  _CREDENTIALS._serialized_start=61
  _CREDENTIALS._serialized_end=147
  _HEADER._serialized_start=149
  _HEADER._serialized_end=185
  _ENDPOINT._serialized_start=188
  _ENDPOINT._serialized_end=338
# @@protoc_insertion_point(module_scope)



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import logging

from bumble.transport.common import StreamPacketSink, StreamPacketSource, Transport

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
async def open_file_transport(spec: str) -> Transport:
    '''
    Open a File transport (typically not for a real file, but for a PTY or other unix
    virtual files).
    The parameter string is the path of the file to open.
    '''

    # Open the file
    file = open(spec, 'r+b', buffering=0)

    # Setup reading
    read_transport, packet_source = await asyncio.get_running_loop().connect_read_pipe(
        StreamPacketSource, file
    )

    # Setup writing
    write_transport, _ = await asyncio.get_running_loop().connect_write_pipe(
        asyncio.BaseProtocol, file
    )
    packet_sink = StreamPacketSink(write_transport)

    class FileTransport(Transport):
        async def close(self):
            read_transport.close()
            write_transport.close()
            file.close()

    return FileTransport(packet_source, packet_sink)



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import random

import pytest

from bumble.codecs import AacAudioRtpPacket, BitReader, BitWriter


# -----------------------------------------------------------------------------
def test_reader():
    reader = BitReader(b'')
    with pytest.raises(ValueError):
        reader.read(1)

    reader = BitReader(b'hello')
    with pytest.raises(ValueError):
        reader.read(40)

    reader = BitReader(bytes([0xFF]))
    assert reader.read(1) == 1
    with pytest.raises(ValueError):
        reader.read(10)

    reader = BitReader(bytes([0x78]))
    value = 0
    for _ in range(8):
        value = (value << 1) | reader.read(1)
    assert value == 0x78

    data = bytes([x & 0xFF for x in range(66 * 100)])
    reader = BitReader(data)
    value = 0
    for _ in range(100):
        for bits in range(1, 33):
            value = value << bits | reader.read(bits)
    assert value == int.from_bytes(data, byteorder='big')


def test_writer():
    writer = BitWriter()
    assert bytes(writer) == b''

    for i in range(100):
        for j in range(1, 10):
            writer = BitWriter()
            chunks = []
            for k in range(j):
                n_bits = random.randint(1, 32)
                random_bits = random.getrandbits(n_bits)
                chunks.append((n_bits, random_bits))
                writer.write(random_bits, n_bits)

            written_data = bytes(writer)
            reader = BitReader(written_data)
            for n_bits, written_bits in chunks:
                read_bits = reader.read(n_bits)
                assert read_bits == written_bits


def test_aac_rtp():
    # pylint: disable=line-too-long
    packet_data = bytes.fromhex(
        '47fc0000b090800300202066000198000de120000000000000000000000000000000000000000000001c'
    )
    packet = AacAudioRtpPacket.from_bytes(packet_data)
    adts = packet.to_adts()
    assert adts == bytes.fromhex(
        'fff1508004fffc2066000198000de120000000000000000000000000000000000000000000001c'
    )

    payload = bytes(list(range(1, 200)))
    rtp = AacAudioRtpPacket.for_simple_aac(44100, 2, payload)
    assert rtp.audio_mux_element.payload == payload
    assert (
        rtp.audio_mux_element.stream_mux_config.audio_specific_config.sampling_frequency
        == 44100
    )
    assert (
        rtp.audio_mux_element.stream_mux_config.audio_specific_config.channel_configuration
        == 2
    )
    rtp2 = AacAudioRtpPacket.from_bytes(bytes(rtp))
    assert str(rtp2.audio_mux_element.stream_mux_config) == str(
        rtp.audio_mux_element.stream_mux_config
    )
    assert rtp2.audio_mux_element.payload == rtp.audio_mux_element.payload


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    test_reader()
    test_writer()
    test_aac_rtp()



import time

from mobly import base_test, test_runner
from mobly.controllers import android_device


class TwoDevicesBenchTest(base_test.BaseTestClass):
    def setup_class(self):
        self.ads = self.register_controller(android_device)
        self.dut1 = self.ads[0]
        self.dut1.load_snippet("bench", "com.github.google.bumble.btbench")
        self.dut2 = self.ads[1]
        self.dut2.load_snippet("bench", "com.github.google.bumble.btbench")

    def test_rfcomm_client_send_receive(self):
        print("### Starting Receiver")
        receiver = self.dut2.bench.runRfcommServer("receive")
        receiver_id = receiver["id"]
        print("--- Receiver status:", receiver)
        while not receiver["model"]["running"]:
            print("--- Waiting for Receiver to be running...")
            time.sleep(1)
            receiver = self.dut2.bench.getRunner(receiver_id)

        print("### Starting Sender")
        sender = self.dut1.bench.runRfcommClient(
            "send", "DC:E5:5B:E5:51:2C", 100, 970, 100
        )
        print("--- Sender status:", sender)

        print("--- Waiting for Sender to complete...")
        sender_result = self.dut1.bench.waitForRunnerCompletion(sender["id"])
        print("--- Sender result:", sender_result)


if __name__ == "__main__":
    test_runner.main()
